@import 'https://fonts.googleapis.com/css?family=Lato:400,700,900|Open+Sans|Roboto+Mono';

pre, code {
  font-family: "Roboto Mono", monospace;
  font-size: 14px;
}
pre {
  line-height: 1.4em;
  overflow-x: auto;
}

code, tt {
  padding: 1px 3px;
  font-family: "Roboto Mono", monospace;
  font-size: 0.85em;
  white-space: pre-wrap;
}

pre {
  margin: 0 0 1.75em 0;
  width: 100%;
  padding: 10px;
  font-family: "Roboto Mono", monospace;
  font-size: 0.75em;
  white-space: pre;
  overflow: auto;
}

pre code, pre tt {
  font-size: inherit;
  white-space: pre-wrap;
  border: none;
  padding: 0;
}

/* better styling for tables, i.e. pd.DataFrame */

table {
  border: none;
  border-collapse: collapse;
  border-spacing: 0;
  table-layout: fixed;
  margin-left: auto;
  margin-right: auto;
}

thead {
  vertical-align: bottom;
  border-bottom: 1px solid;
  text-align: right;
}

/* alabaster has some weird div.body table defines,
 * and I can't override them by `td, th` or 
 * `.dataframe td`. WTF.
div.body td, div.body th {
  text-align: right;
  vertical-align: right;
  padding: 0.5em 0.5em;
  line-height: 1.0;
  white-space: nowrap;
  max-width: 100px;
  text-overflow: ellipsis;
  overflow: hidden;
  border: none;
}
 */

div.sphinxsidebar ul li {
  line-height: 1.8;
}
