# coding=utf-8
# Filename: __init__.py
"""
A collection of utils.

"""
from __future__ import division, absolute_import, print_function

from km3pipe.utils import h5tree, h5info, tohdf5, h5concat         # noqa

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"
